// -------------------------------------------------------------------------
// File:    reCamera.cxx
// Desc:    base class for a camera (position, view direction, 
//          upvec, FOV, viewport, etc)
//
// Author:  Tomas Mller
// History: March, 2000 (started)
// -------------------------------------------------------------------------

#include <GL/glut.h>

#include "misc.h"
#include "reCamera.h"

reCamera::reCamera()
{
   mPosition.set(0,0,5);
   mDirection.set(0,0,-1);
   mLookAt.set(0,0,0);
   mUpVector.set(0,1,0);
   mVFOV=45.0;
   mWidth=256;
   mHeight=256;
   mNear=0.1f;
   mFar=100.0f;
   mPreComputed=false;

   precomputeTracerParams();
}

reCamera::~reCamera()
{
}

void reCamera::setPosition( const Vec3f pos )
{
  mPosition = pos ;
  mDirection = mLookAt - mPosition ;  // compute new direction
  mDirection.normalize() ;
  mPreComputed = false ;
}


void reCamera::setUpVector(const Vec3f up)
{
   mUpVector=up;
   mUpVector.normalize();
   ASSERT(mUpVector*mUpVector>fEPSILON);
   mPreComputed=false;
}

void reCamera::setLookAt(const Vec3f lookat)
{
   mLookAt=lookat;
   mDirection=mLookAt-mPosition;  // compute new direction
   mDirection.normalize();
}


void reCamera::setVFOV(float vfov)
{
   mVFOV=vfov;
   mPreComputed=false;
}

void reCamera::setResolution(int width,int height)
{
   mWidth=width;
   mHeight=height;
   mPreComputed=false;
}

Vec3f reCamera::getPosition(void) const
{
   return mPosition;
}

Vec3f reCamera::getDirection(void) const
{
   return mDirection;
}

Vec3f reCamera::getUpVector(void) const
{
   return mUpVector;
}

Vec3f reCamera::getLookAt(void) const
{
   return mLookAt;
}

float reCamera::getVFOV(void) const
{
   return mVFOV;
}

int reCamera::getWidth(void) const
{
   return mWidth;
}

int reCamera::getHeight(void) const
{
   return mHeight;
}

void reCamera::setNearFar(float near_value,float far_value)
{
   mNear=near_value;
   mFar=far_value;
}

void reCamera::getNearFar(float &near_value,float &far_value) const
{
   near_value=mNear;
   far_value=mFar;
}

float reCamera::getNear(void) const
{
   return mNear;
}

float reCamera::getFar(void) const
{
   return mFar;
}

void reCamera::precomputeTracerParams(void)
{
   float height,width;
   Vec3f xdir,ydir;

   xdir=mDirection%mUpVector;
   xdir.normalize();
   ydir=xdir%mDirection;
   ydir.normalize();

   height=2.0*tan(mVFOV*TORAD/2.0);
   ASSERT(mHeight>0);
   mDeltaY=ydir*(height/mHeight);

   width=2.0*tan(((float)mWidth/mHeight)*mVFOV*TORAD/2.0);
   ASSERT(mWidth>0);
   mDeltaX=xdir*(width/mWidth);
   
   mLowerLeft=mPosition+mDirection-
      mDeltaX*((mWidth-1.0)/2.0)-
      mDeltaY*((mHeight-1.0)/2.0);

   mPreComputed=true;
}

void reCamera::getTracerParams(Vec3f &campos, Vec3f &lower_left,
			       Vec3f &deltax,Vec3f &deltay)
{
   if(!mPreComputed)
   {
      precomputeTracerParams();
   }
   campos=mPosition;
   lower_left=mLowerLeft;
   deltax=mDeltaX;
   deltay=mDeltaY;
}

void reCamera::setOGLMatrices(void) const  // set up OpenGL matrices
{
//   glViewport(0, 0, mWidth, mHeight);
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective((real)mVFOV,
                  (real)mWidth/mHeight, // aspect ratio: is this the right way to do it?
                  (real)mNear,(real)mFar);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(mPosition[0],mPosition[1],mPosition[2],
			 mLookAt[0],mLookAt[1],mLookAt[2],
             mUpVector[0],mUpVector[1],mUpVector[2]);
}


void reCamera::renderOGL(void) const
{
  Vec3r p[8];
  Vec3r delta_u,delta_v,distvec;
  glDisable(GL_LIGHTING);
  glColor3f(1,1,1);

  Vec3r right=mUpVector%mDirection;
  
  glBegin(GL_LINES);
  glVertex3fv(mPosition.vec);
  glVertex3fv(mLookAt.vec);
  glEnd();  
  
  distvec=mPosition+mDirection*mNear;
  delta_v=mUpVector*tan(mVFOV*TORAD*0.5)*mNear;
  delta_u=right*(tan(mVFOV*TORAD*0.5)*(float)mWidth/mHeight*mNear);
  
  
  p[0]=distvec - delta_u - delta_v;
  p[1]=distvec + delta_u - delta_v;
  p[2]=distvec + delta_u + delta_v;
  p[3]=distvec - delta_u + delta_v;
  glBegin(GL_LINE_STRIP);
  glVertex3fv(p[0].vec);
  glVertex3fv(p[1].vec);
  glVertex3fv(p[2].vec);
  glVertex3fv(p[3].vec);
  glVertex3fv(p[0].vec);
  glEnd();
  
  distvec=mPosition+mDirection*mFar;
  delta_v=mUpVector*tan(mVFOV*TORAD*0.5)*mFar;
  delta_u=right*(tan(mVFOV*TORAD*0.5)*(float)mWidth/mHeight*mFar);
  
  p[4]=distvec - delta_u - delta_v;
  p[5]=distvec + delta_u - delta_v;
  p[6]=distvec + delta_u + delta_v;
  p[7]=distvec - delta_u + delta_v;
  glBegin(GL_LINE_STRIP);
  glVertex3fv(p[4].vec);
  glVertex3fv(p[5].vec);
  glVertex3fv(p[6].vec);
  glVertex3fv(p[7].vec);
  glVertex3fv(p[4].vec);
  glEnd();
  
  glBegin(GL_LINES);
  glVertex3fv(mPosition.vec);
  glVertex3fv(p[4].vec);
  glVertex3fv(mPosition.vec);
  glVertex3fv(p[5].vec);
  glVertex3fv(mPosition.vec);
  glVertex3fv(p[6].vec);
  glVertex3fv(mPosition.vec);
  glVertex3fv(p[7].vec);
  
  glEnd();
  
  glEnable(GL_LIGHTING);
}

void reCamera::debugprint(void)
{
   printf("Camera:\n");
   printf(" from: "); mPosition.debugprint();
   printf("   at: "); mLookAt.debugprint();
   printf("  dir: "); mDirection.debugprint();
   printf("   up: "); mUpVector.debugprint();

   printf(" near: %f\n",mNear);
   printf("  far: %f\n",mFar);
   printf(" VFOV: %f\n",mVFOV);
   printf("  res: %d x %d\n",mWidth,mHeight);
}

